@@ -24,7 +24,6 @@ export PKG_CONFIG_PATH="$inst/lib/pkgconfig:$PKG_CONFIG_PATH"
 export LD_LIBRARY_PATH="$inst/lib"
 
 osmo-build-dep.sh libosmocore "" --disable-doxygen
-osmo-build-dep.sh libosmo-abis
 
 set +x
 echo
@@ -35,10 +34,10 @@ echo
 set -x
 
 autoreconf --install --force
-./configure --enable-sanitize --enable-werror
+./configure --enable-sanitize --enable-werror --enable-lapd-examples
 $MAKE $PARALLEL_MAKE
-$MAKE distcheck \
+$MAKE $PARALLEL_MAKE distcheck \
   || cat-testlogs.sh
-$MAKE maintainer-clean
+$MAKE $PARALLEL_MAKE maintainer-clean
 
 osmo-clean-workspace.sh
