@@ -74,7 +74,7 @@ static int connect_cb(struct osmo_stream_cli *conn)
 		char *ptr;
 		int x;
 
-		msg = osmo_ipa_msg_alloc(0);
+		msg = osmo_ipa_ext_msg_alloc(0);
 		if (msg == NULL) {
 			LOGP(DLINP, LOGL_ERROR, "cannot alloc msg\n");
 			return -1;
@@ -93,8 +93,7 @@ static int connect_cb(struct osmo_stream_cli *conn)
 		msg_sent->num = i;
 		llist_add(&msg_sent->head, &msg_sent_list);
 
-		osmo_ipa_msg_push_header(msg, IPAC_PROTO_OSMO);
-
+		osmo_ipa_msg_push_headers(msg, osmo_ipa_msgb_cb_proto(msg), osmo_ipa_msgb_cb_proto_ext(msg));
 		osmo_stream_cli_send(conn, msg);
 
 		LOGP(DIPATEST, LOGL_DEBUG, "enqueueing msg %d of "
@@ -103,30 +102,20 @@ static int connect_cb(struct osmo_stream_cli *conn)
 	return 0;
 }
 
-static int read_cb(struct osmo_stream_cli *conn)
+static int read_cb(struct osmo_stream_cli *conn, int res, struct msgb *msg)
 {
-	struct msgb *msg;
-
-	LOGP(DIPATEST, LOGL_DEBUG, "received message from stream\n");
-
-	msg = osmo_ipa_msg_alloc(0);
-	if (msg == NULL) {
-		LOGP(DIPATEST, LOGL_ERROR, "cannot allocate message\n");
-		return 0;
-	}
-	if (osmo_stream_cli_recv(conn, msg) <= 0) {
-		LOGP(DIPATEST, LOGL_ERROR, "cannot receive message\n");
-		return 0;
-	}
-	if (osmo_ipa_process_msg(msg) < 0) {
-		LOGP(DIPATEST, LOGL_ERROR, "bad IPA message\n");
-		return 0;
-	}
-
 	int num;
 	struct msg_sent *cur, *tmp, *found = NULL;
 
-	num = ntohl(*((int *)(msg->data + sizeof(struct ipa_head))));
+	LOGP(DIPATEST, LOGL_DEBUG, "received message from stream (payload len=%d)\n", msgb_length(msg));
+
+	if (res <= 0) {
+		LOGP(DIPATEST, LOGL_ERROR, "Event with no data! %d\n", res);
+		msgb_free(msg);
+		return 0;
+	}
+
+	num = osmo_load32be(msgb_l2(msg));
 	LOGP(DLINP, LOGL_DEBUG, "received msg number %d\n", num);
 
 	llist_for_each_entry_safe(cur, tmp, &msg_sent_list, head) {
@@ -149,6 +138,7 @@ static int read_cb(struct osmo_stream_cli *conn)
 		LOGP(DLINP, LOGL_ERROR,
 			"message %d not found!\n", num);
 	}
+	msgb_free(msg);
 	return 0;
 }
 
@@ -169,7 +159,7 @@ int main(int argc, char *argv[])
 	tall_test = talloc_named_const(NULL, 1, "osmo_stream_client_test");
 	msgb_talloc_ctx_init(tall_test, 0);
 	osmo_init_logging2(tall_test, &osmo_stream_client_test_log_info);
-	log_set_log_level(osmo_stderr_target, LOGL_NOTICE);
+	log_set_log_level(osmo_stderr_target, LOGL_DEBUG);
 
 	/*
 	 * initialize stream client.
@@ -180,11 +170,12 @@ int main(int argc, char *argv[])
 		fprintf(stderr, "cannot create client\n");
 		exit(EXIT_FAILURE);
 	}
+	osmo_stream_cli_set_name(conn, "ipa_test_client");
 	osmo_stream_cli_set_addr(conn, "127.0.0.1");
 	osmo_stream_cli_set_port(conn, 10000);
 	osmo_stream_cli_set_connect_cb(conn, connect_cb);
 	osmo_stream_cli_set_disconnect_cb(conn, disconnect_cb);
-	osmo_stream_cli_set_read_cb(conn, read_cb);
+	osmo_stream_cli_set_read_cb2(conn, read_cb);
 	osmo_stream_cli_set_data(conn, &num_msgs);
 	osmo_stream_cli_set_nodelay(conn, true);
 
@@ -193,6 +184,8 @@ int main(int argc, char *argv[])
 		exit(EXIT_FAILURE);
 	}
 
+	osmo_stream_cli_set_segmentation_cb(conn, osmo_ipa_segmentation_cb);
+
 	LOGP(DIPATEST, LOGL_NOTICE, "Entering main loop\n");
 
 	while(1) {
