@@ -62,7 +62,7 @@ void rs232_tx_timer_cb(void *ptr)
 	struct osmo_rs232 *r = ptr;
 
 	/* we're again ready to transmit. */
-	r->ofd.when |= OSMO_FD_WRITE;
+	osmo_fd_write_enable(&r->ofd);
 }
 
 static int handle_ser_write(struct osmo_fd *bfd)
@@ -72,15 +72,15 @@ static int handle_ser_write(struct osmo_fd *bfd)
 	struct msgb *msg;
 	int written;
 
-        LOGP(DLINP, LOGL_DEBUG, "writing data to rs232\n");
+	LOGP(DLINP, LOGL_DEBUG, "writing data to rs232\n");
 
-        if (llist_empty(&r->tx_queue)) {
-                r->ofd.when &= ~OSMO_FD_WRITE;
-                return 0;
-        }
-        lh = r->tx_queue.next;
-        llist_del(lh);
-        msg = llist_entry(lh, struct msgb, list);
+	if (llist_empty(&r->tx_queue)) {
+		osmo_fd_write_disable(&r->ofd);
+		return 0;
+	}
+	lh = r->tx_queue.next;
+	llist_del(lh);
+	msg = llist_entry(lh, struct msgb, list);
 
 	written = write(bfd->fd, msg->data, msg->len);
 	if (written < msg->len) {
@@ -92,7 +92,7 @@ static int handle_ser_write(struct osmo_fd *bfd)
 
 	/* We've got more data to write, but we have to wait to make it. */
 	if (!llist_empty(&r->tx_queue) && r->cfg.delay_us) {
-		r->ofd.when &= ~OSMO_FD_WRITE;
+		osmo_fd_write_disable(&r->ofd);
 		osmo_timer_schedule(&r->tx_timer, 0, r->cfg.delay_us);
 	}
 	return 0;
@@ -223,9 +223,7 @@ int osmo_rs232_open(struct osmo_rs232 *r)
 		return rc;
 	}
 
-	bfd->when = OSMO_FD_READ;
-	bfd->cb = serial_fd_cb;
-	bfd->data = r;
+	osmo_fd_setup(bfd, bfd->fd, OSMO_FD_READ, serial_fd_cb, r, 0);
 
 	rc = osmo_fd_register(bfd);
 	if (rc < 0) {
@@ -256,8 +254,8 @@ int osmo_rs232_read(struct osmo_rs232 *r, struct msgb *msg)
 
 void osmo_rs232_write(struct osmo_rs232 *r, struct msgb *msg)
 {
-        msgb_enqueue(&r->tx_queue, msg);
-        r->ofd.when |= OSMO_FD_WRITE;
+	msgb_enqueue(&r->tx_queue, msg);
+	osmo_fd_write_enable(&r->ofd);
 }
 
 void osmo_rs232_close(struct osmo_rs232 *r)
