@@ -117,6 +117,7 @@ struct osmo_stream_cli {
 	osmo_stream_cli_read_cb_t	read_cb;
 	osmo_stream_cli_read_cb2_t	iofd_read_cb;
 	osmo_stream_cli_segmentation_cb_t segmentation_cb;
+	osmo_stream_cli_segmentation_cb2_t segmentation_cb2;
 	void				*data;
 	int				flags;
 	int				reconnect_timeout;
@@ -203,6 +204,7 @@ static bool stream_cli_close(struct osmo_stream_cli *cli)
 	 * Also, if reconnect is disabled by user, notify the user that connect() failed: */
 	if (old_state == STREAM_CLI_STATE_CONNECTED ||
 	    (old_state == STREAM_CLI_STATE_CONNECTING && cli->reconnect_timeout < 0)) {
+		OSMO_ASSERT(!(cli->in_cb_mask & IN_CB_MASK_DISCONNECT_CB));
 		cli->in_cb_mask |= IN_CB_MASK_DISCONNECT_CB;
 		if (cli->disconnect_cb)
 			cli->disconnect_cb(cli);
@@ -218,7 +220,7 @@ static bool stream_cli_close(struct osmo_stream_cli *cli)
  *  abstraction and close the socket */
 void osmo_stream_cli_close(struct osmo_stream_cli *cli)
 {
-	stream_cli_close(cli);
+	(void)stream_cli_close(cli);
 }
 
 /*! Re-connect an Osmocom Stream Client.
@@ -250,7 +252,7 @@ static bool stream_cli_reconnect(struct osmo_stream_cli *cli)
  *  we close any existing connection (if any) and schedule a re-connect timer */
 void osmo_stream_cli_reconnect(struct osmo_stream_cli *cli)
 {
-	stream_cli_reconnect(cli);
+	(void)stream_cli_reconnect(cli);
 }
 
 /*! Check if Osmocom Stream Client is in connected state.
@@ -382,7 +384,7 @@ static int stream_cli_write(struct osmo_stream_cli *cli)
 			return 0;
 		}
 		msgb_free(msg);
-		stream_cli_reconnect(cli);
+		(void)stream_cli_reconnect(cli);
 		return 0;
 	}
 
@@ -408,7 +410,8 @@ static int _setsockopt_nosigpipe(struct osmo_stream_cli *cli)
 #endif
 }
 
-static void stream_cli_handle_connecting(struct osmo_stream_cli *cli, int res)
+/* returns true if cli is freed */
+static bool stream_cli_handle_connecting(struct osmo_stream_cli *cli, int res)
 {
 	int error, ret = res;
 	socklen_t len = sizeof(error);
@@ -418,14 +421,12 @@ static void stream_cli_handle_connecting(struct osmo_stream_cli *cli, int res)
 
 	if (ret < 0) {
 		LOGSCLI(cli, LOGL_ERROR, "connect failed (%d)\n", res);
-		stream_cli_reconnect(cli);
-		return;
+		return stream_cli_reconnect(cli);
 	}
 	ret = getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &len);
 	if (ret >= 0 && error > 0) {
 		LOGSCLI(cli, LOGL_ERROR, "connect so_error (%d)\n", error);
-		stream_cli_reconnect(cli);
-		return;
+		return stream_cli_reconnect(cli);
 	}
 
 	/* If messages got enqueued while 'connecting', keep WRITE flag
@@ -457,7 +458,7 @@ static void stream_cli_handle_connecting(struct osmo_stream_cli *cli, int res)
 	if (cli->connect_cb)
 		cli->connect_cb(cli);
 	cli->in_cb_mask &= ~IN_CB_MASK_CONNECT_CB;
-	free_delayed_if_needed(cli);
+	return free_delayed_if_needed(cli);
 }
 
 static int osmo_stream_cli_fd_cb(struct osmo_fd *ofd, unsigned int what)
@@ -466,7 +467,7 @@ static int osmo_stream_cli_fd_cb(struct osmo_fd *ofd, unsigned int what)
 
 	switch (cli->state) {
 	case STREAM_CLI_STATE_CONNECTING:
-		stream_cli_handle_connecting(cli, 0);
+		(void)stream_cli_handle_connecting(cli, 0);
 		break;
 	case STREAM_CLI_STATE_CONNECTED:
 		if (what & OSMO_FD_READ) {
@@ -511,7 +512,6 @@ struct osmo_stream_cli *osmo_stream_cli_create(void *ctx)
 	cli->state = STREAM_CLI_STATE_CLOSED;
 	osmo_timer_setup(&cli->timer, cli_timer_cb, cli);
 	cli->reconnect_timeout = 5;	/* default is 5 seconds. */
-	cli->segmentation_cb = NULL;
 	INIT_LLIST_HEAD(&cli->tx_queue);
 	cli->tx_queue_max_length = 1024; /* Default tx queue size, msgbs. */
 
@@ -527,8 +527,14 @@ static void stream_cli_iofd_read_cb(struct osmo_io_fd *iofd, int res, struct msg
 
 	switch (cli->state) {
 	case STREAM_CLI_STATE_CONNECTING:
-		msgb_free(msg);
-		stream_cli_handle_connecting(cli, res);
+		freed = stream_cli_handle_connecting(cli, res);
+		if (freed)
+			return; /* msg was also freed as part of the talloc tree. */
+		if (cli->state != STREAM_CLI_STATE_CONNECTED) {
+			msgb_free(msg);
+			return;
+		}
+		/* Follow below common path submitting read_cb(msg) to user. */
 		break;
 	case STREAM_CLI_STATE_CONNECTED:
 		switch (res) {
@@ -548,35 +554,41 @@ static void stream_cli_iofd_read_cb(struct osmo_io_fd *iofd, int res, struct msg
 		}
 		if (freed)
 			return; /* msg was also freed as part of the talloc tree. */
-		/* Notify user of new data or error: */
-		if (!cli->iofd_read_cb) {
-			msgb_free(msg);
-			return;
-		}
-		cli->in_cb_mask |= IN_CB_MASK_READ_CB;
-		cli->iofd_read_cb(cli, res, msg);
-		cli->in_cb_mask &= ~IN_CB_MASK_READ_CB;
-		OSMO_ASSERT(cli->in_cb_mask == 0);
-		free_delayed_if_needed(cli);
+		/* Follow below common path submitting read_cb(msg) to user. */
 		break;
 	default:
 		osmo_panic("%s() called with unexpected state %d\n", __func__, cli->state);
 	}
+
+	/* Notify user of new data or error: */
+	if (!cli->iofd_read_cb) {
+		msgb_free(msg);
+		return;
+	}
+	cli->in_cb_mask |= IN_CB_MASK_READ_CB;
+	cli->iofd_read_cb(cli, res, msg);
+	cli->in_cb_mask &= ~IN_CB_MASK_READ_CB;
+	OSMO_ASSERT(cli->in_cb_mask == 0);
+	(void)free_delayed_if_needed(cli);
 }
 
 static void stream_cli_iofd_write_cb(struct osmo_io_fd *iofd, int res, struct msgb *msg)
 {
 	struct osmo_stream_cli *cli = osmo_iofd_get_data(iofd);
+	/* msgb is not owned by us here, no need to free it. */
 
 	switch (cli->state) {
 	case STREAM_CLI_STATE_CONNECTING:
-		stream_cli_handle_connecting(cli, res);
+		(void)stream_cli_handle_connecting(cli, res);
 		break;
 	case STREAM_CLI_STATE_CONNECTED:
 		if (msg && res <= 0) {
 			LOGSCLI(cli, LOGL_ERROR, "received error %d in response to send\n", res);
-			stream_cli_reconnect(cli);
+			(void)stream_cli_reconnect(cli);
 		}
+		/* res=0 && msgb=NULL: "connected notify", but we already received before a read_cb
+		 * which moved us to CONNECTED state. Do nothing.
+		 */
 		break;
 	default:
 		osmo_panic("%s() called with unexpected state %d\n", __func__, cli->state);
@@ -586,8 +598,6 @@ static void stream_cli_iofd_write_cb(struct osmo_io_fd *iofd, int res, struct ms
 static const struct osmo_io_ops osmo_stream_cli_ioops = {
 	.read_cb = stream_cli_iofd_read_cb,
 	.write_cb = stream_cli_iofd_write_cb,
-
-	.segmentation_cb = NULL,
 };
 
 #ifdef HAVE_LIBSCTP
@@ -600,8 +610,14 @@ static void stream_cli_iofd_recvmsg_cb(struct osmo_io_fd *iofd, int res, struct
 
 	switch (cli->state) {
 	case STREAM_CLI_STATE_CONNECTING:
-		msgb_free(msg);
-		stream_cli_handle_connecting(cli, res);
+		freed = stream_cli_handle_connecting(cli, res);
+		if (freed)
+			return; /* msg was also freed as part of the talloc tree. */
+		if (cli->state != STREAM_CLI_STATE_CONNECTED) {
+			msgb_free(msg);
+			return;
+		}
+		/* Follow below common path submitting read_cb(msg) to user. */
 		break;
 	case STREAM_CLI_STATE_CONNECTED:
 		switch (res) {
@@ -620,27 +636,27 @@ static void stream_cli_iofd_recvmsg_cb(struct osmo_io_fd *iofd, int res, struct
 		}
 		if (freed)
 			return; /* msg was also freed as part of the talloc tree. */
-		/* Notify user of new data or error: */
-		if (!cli->iofd_read_cb) {
-			msgb_free(msg);
-			return;
-		}
-		cli->in_cb_mask |= IN_CB_MASK_READ_CB;
-		cli->iofd_read_cb(cli, res, msg);
-		cli->in_cb_mask &= ~IN_CB_MASK_READ_CB;
-		OSMO_ASSERT(cli->in_cb_mask == 0);
-		free_delayed_if_needed(cli);
+		/* Follow below common path submitting read_cb(msg) to user. */
 		break;
 	default:
 		osmo_panic("%s() called with unexpected state %d\n", __func__, cli->state);
 	}
+
+	/* Notify user of new data or error: */
+	if (!cli->iofd_read_cb) {
+		msgb_free(msg);
+		return;
+	}
+	cli->in_cb_mask |= IN_CB_MASK_READ_CB;
+	cli->iofd_read_cb(cli, res, msg);
+	cli->in_cb_mask &= ~IN_CB_MASK_READ_CB;
+	OSMO_ASSERT(cli->in_cb_mask == 0);
+	(void)free_delayed_if_needed(cli);
 }
 
 static const struct osmo_io_ops osmo_stream_cli_ioops_sctp = {
 	.recvmsg_cb = stream_cli_iofd_recvmsg_cb,
 	.sendmsg_cb = stream_cli_iofd_write_cb,
-
-	.segmentation_cb = NULL,
 };
 #endif
 
@@ -775,28 +791,66 @@ osmo_stream_cli_set_proto(struct osmo_stream_cli *cli, uint16_t proto)
 	cli->flags |= OSMO_STREAM_CLI_F_RECONF;
 }
 
+/* Callback from iofd, forward to stream_cli user: */
+static int stream_cli_iofd_segmentation_cb2(struct osmo_io_fd *iofd, struct msgb *msg)
+{
+	struct osmo_stream_cli *cli = osmo_iofd_get_data(iofd);
+	if (cli->segmentation_cb2)
+		return cli->segmentation_cb2(cli, msg);
+	if (cli->segmentation_cb)
+		return cli->segmentation_cb(msg);
+	OSMO_ASSERT(0);
+	return 0;
+}
+
 /* Configure client side segmentation for the iofd */
-static void configure_cli_segmentation_cb(struct osmo_stream_cli *cli,
-					  osmo_stream_cli_segmentation_cb_t segmentation_cb)
+static void configure_cli_segmentation_cb(struct osmo_stream_cli *cli)
 {
 	/* Copy default settings */
 	struct osmo_io_ops client_ops;
 	osmo_iofd_get_ioops(cli->iofd, &client_ops);
 	/* Set segmentation cb for this client */
-	client_ops.segmentation_cb = segmentation_cb;
+	if (cli->segmentation_cb || cli->segmentation_cb2)
+		client_ops.segmentation_cb2 = stream_cli_iofd_segmentation_cb2;
+	else
+		client_ops.segmentation_cb2 = NULL;
 	osmo_iofd_set_ioops(cli->iofd, &client_ops);
 }
 
 /*! Set the segmentation callback for the client.
  *  \param[in,out] cli Stream Client to modify
  *  \param[in] segmentation_cb Target segmentation callback
+ *
+ *  A segmentation call-back can optionally be used when a packet based protocol
+ *  (like TCP) is used within a STREAM style socket that does not preserve
+ *  message boundaries within the stream.  If a segmentation call-back is given,
+ *  the osmo_stream_srv library code will makes sure that the read_cb called
+ *  only for complete single messages, and not arbitrary segments of the stream.
  */
 void osmo_stream_cli_set_segmentation_cb(struct osmo_stream_cli *cli,
 					 osmo_stream_cli_segmentation_cb_t segmentation_cb)
 {
 	cli->segmentation_cb = segmentation_cb;
+	cli->segmentation_cb2 = NULL;
 	if (cli->iofd) /* Otherwise, this will be done in osmo_stream_cli_open() */
-		configure_cli_segmentation_cb(cli, segmentation_cb);
+		configure_cli_segmentation_cb(cli);
+}
+
+/*! Set the segmentation callback for the client.
+ *  \param[in,out] cli Stream Client to modify
+ *  \param[in] segmentation_cb2 Target segmentation callback
+ *
+ * Same as osmo_stream_cli_set_segmentation_cb(), but a
+ * osmo_stream_cli_segmentation_cb2_t is called instead which allows access to
+ * the related cli object.
+ */
+void osmo_stream_cli_set_segmentation_cb2(struct osmo_stream_cli *cli,
+					  osmo_stream_cli_segmentation_cb2_t segmentation_cb2)
+{
+	cli->segmentation_cb = NULL;
+	cli->segmentation_cb2 = segmentation_cb2;
+	if (cli->iofd) /* Otherwise, this will be done in osmo_stream_cli_open() */
+		configure_cli_segmentation_cb(cli);
 }
 
 /*! Set the socket type for the stream server link.
@@ -966,7 +1020,7 @@ void osmo_stream_cli_destroy(struct osmo_stream_cli *cli)
 		return;
 
 	LOGSCLI(cli, LOGL_DEBUG, "destroy()\n");
-	OSMO_ASSERT(!stream_cli_close(cli));
+	OSMO_ASSERT(stream_cli_close(cli) == false);
 	osmo_timer_del(&cli->timer);
 	msgb_queue_free(&cli->tx_queue);
 	cli->tx_queue_count = 0;
@@ -1025,7 +1079,7 @@ int osmo_stream_cli_open2(struct osmo_stream_cli *cli, int reconnect)
 	if (ret < 0) {
 		LOGSCLI(cli, LOGL_ERROR, "connect: socket creation error (%d)\n", ret);
 		if (reconnect)
-			stream_cli_reconnect(cli);
+			(void)stream_cli_reconnect(cli);
 		return ret;
 	}
 	osmo_fd_setup(&cli->ofd, ret, OSMO_FD_READ | OSMO_FD_WRITE, osmo_stream_cli_fd_cb, cli, 0);
@@ -1198,7 +1252,7 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 
 	if (ret < 0) {
 		LOGSCLI(cli, LOGL_ERROR, "connect: socket creation error (%d)\n", ret);
-		stream_cli_reconnect(cli);
+		(void)stream_cli_reconnect(cli);
 		return ret;
 	}
 
@@ -1237,7 +1291,7 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 
 		osmo_iofd_set_txqueue_max_length(cli->iofd, cli->tx_queue_max_length);
 		osmo_iofd_notify_connected(cli->iofd);
-		configure_cli_segmentation_cb(cli, cli->segmentation_cb);
+		configure_cli_segmentation_cb(cli);
 
 		if (osmo_iofd_register(cli->iofd, fd) < 0)
 			goto error_close_socket;
@@ -1367,11 +1421,11 @@ int osmo_stream_cli_recv(struct osmo_stream_cli *cli, struct msgb *msg)
 			LOGSCLI(cli, LOGL_ERROR, "lost connection with srv (%d)\n", errno);
 		else
 			LOGSCLI(cli, LOGL_ERROR, "recv failed (%d)\n", errno);
-		stream_cli_reconnect(cli);
+		(void)stream_cli_reconnect(cli);
 		return ret;
 	} else if (ret == 0) {
 		LOGSCLI(cli, LOGL_ERROR, "connection closed with srv\n");
-		osmo_stream_cli_reconnect(cli);
+		(void)stream_cli_reconnect(cli);
 		return ret;
 	}
 	msgb_put(msg, ret);
